#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <windows.h>

/****************************************/

/* almost all values are path */
#define VALUE_SIZE (PATH_MAX + 1 /* for '\0' */)
/* expected not so long */
#define ENV_NAME_MAX 32

/* this includes slot of termination(NULL) */
#define ADDENDUM (sizeof(requireds) / sizeof(*requireds))

/********************/

#define SHOW_UNIXERROR(message, hint) \
        fprintf(stderr, \
                "ERROR: %s: %s: %s\n", \
                message, hint, (0 != errno ? strerror(errno) : " "))

#define ENSURE_OR_EPILOGUE_E(exp, hint) \
    if(!(exp)){ \
        SHOW_UNIXERROR(#exp, hint); \
        goto epilogue; \
    }


#define ENSURE_OR_EPILOGUE(exp, hint) \
    if(!(exp)){ \
        fprintf(stderr, \
                "ERROR: %s: %s\n", \
                #exp, hint); \
        goto epilogue; \
    }


/********************/

#define SHOW_WIN_ERROR(code, message, hint) \
    fprintf(stderr, \
            "ERROR: %s: %s: %d\n", \
            message, hint, code)

#define ENSURE_SUCCESS_OR_EPILOGUE(result, method, hint) \
    if(ERROR_SUCCESS != result){ \
        SHOW_WIN_ERROR(result, method, hint); \
        goto epilogue; \
    }

/****************************************/

struct addendum
{
    const char* name;
    char value[ENV_NAME_MAX + 1 /* for '=' */ + VALUE_SIZE];
};

static struct addendum requireds[] =
{
    { .name = "SSH_AGENT_PID"  },
    { .name = "SSH_AUTH_SOCK" },
    { .name = "SSH_ASKPASS" },
    { .name = "DISPLAY" },
    { .name = NULL }
};

/*****************************************************************************/

static int 
setup_requireds()
{
    int code = -1; /* of this function */
    LONG result; /* of Win32 API */

    /********************/

    HKEY key = NULL;

    /********************/

    int i;

    /********************/

    {
        const char* name = "Environment";

        result =
        RegOpenKeyEx(HKEY_CURRENT_USER,
                     name,
                     0, /* reserved */
                     KEY_QUERY_VALUE,
                     &key);
        ENSURE_SUCCESS_OR_EPILOGUE(result, "RegOpenKeyEx", name);
    }

    for(i = 0 ; NULL != requireds[i].name ; i += 1){
        const char* name = requireds[i].name;
        char buff[VALUE_SIZE];
        DWORD length = sizeof(buff);

        result =
        RegQueryValueEx(key,
                        name,
                        NULL, /* reserved */
                        NULL, /* type[out] */
                        (BYTE*)buff,
                        &length);

        ENSURE_SUCCESS_OR_EPILOGUE(result, "RegQueryValueEx", name);

        ENSURE_OR_EPILOGUE((strlen(name) < ENV_NAME_MAX), name);
        sprintf(requireds[i].value, "%s=%s", name, buff);
    }

    code = 0; /* means success */

    /********************/
  epilogue:
    if(NULL != key){ RegCloseKey(key); }

    /********************/

    return code;
}

/****************************************/

int
count_environ()
{
    int i;
    for(i = 0 ; NULL != environ[i] ; i += 1){ }
    return i;
}

int
find_in_requireds(const char* envline)
{
    int i;

    for(i = 0 ; NULL != requireds[i].name ; i += 1){
        const char* name = requireds[i].name;
        int length = strlen(name);

        if((0 == strncmp(name, envline, length)) && ('=' == envline[length])){
            return 1;
        }
    }

    return 0;
}

char**
mix_requireds()
{
    char** mixed = NULL;

    /********************/

    int i;
    int index;

    /********************/

    mixed = malloc(sizeof(char*) * (count_environ() + ADDENDUM));
    ENSURE_OR_EPILOGUE_E((NULL != mixed), "malloc");

    index = -1;

    for(i = 0 ; NULL != environ[i] ; i += 1){
        if(find_in_requireds(environ[i])){
            /* skip old value */
            continue;
        }

        mixed[(index += 1)] = environ[i];
    }

    for(i = 0 ; NULL != requireds[i].name ; i += 1){
        mixed[(index += 1)] = requireds[i].value;
    }

    mixed[(index += 1)] = NULL;

    /********************/
  epilogue:
    /* no resource to be released */

    /********************/

    return mixed;
}

/****************************************/

int
main(int argc,
     const char* argv[])
{
    int result = 1; /* means error */

    /********************/

    char** mixed = NULL;

    /********************/

    ENSURE_OR_EPILOGUE((1 < argc), "no executable is specified");

    ENSURE_OR_EPILOGUE((0 == setup_requireds()),
                       "environment variable not found");

    mixed = mix_requireds();
    ENSURE_OR_EPILOGUE((NULL != mixed), "");

    argv[0] = WRAPT_PATH;
    execve(argv[0], (char* const*)argv, mixed);
    /*
     * above does not return control to here, if "execv" successes.
     */
    SHOW_UNIXERROR("execve", argv[0]);

    /********************/
  epilogue:
    if(NULL != mixed){ free(mixed); }

    /********************/

    return result;
}

/*****************************************************************************/
